@@ -8,8 +8,8 @@
 # Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
 # 2006, 2007, 2008, 2009, 2010, 2011 Comite Reseau des Universites
 # Copyright (c) 2011, 2012, 2013, 2014, 2015, 2016, 2017 GIP RENATER
-# Copyright 2018 The Sympa Community. See the AUTHORS.md file at the
-# top-level directory of this distribution and at
+# Copyright 2018, 2021 The Sympa Community. See the
+# AUTHORS.md file at the top-level directory of this distribution and at
 # <https://github.com/sympa-community/sympa.git>.
 #
 # This program is free software; you can redistribute it and/or modify
@@ -63,6 +63,8 @@ sub connect {
     } else {
         $self->__dbh->func(5000, 'busy_timeout');
     }
+    # Create a temoprarhy view "dual" for portable SQL statements.
+    $self->__dbh->do(q{CREATE TEMPORARY VIEW dual AS SELECT 'X' AS dummy;});
 
     return 1;
 }
@@ -288,7 +290,7 @@ sub add_field {
                 q{ALTER TABLE %s ADD %s %s%s},
                 $table, $field, $param->{'type'}, $options
             )
-            ) {
+        ) {
             $log->syslog('err',
                 'Could not add field %s to table %s in database %s',
                 $field, $table, $self->{'db_name'});
@@ -449,7 +451,8 @@ sub unset_index {
         $param->{'index'}, $param->{'table'});
 
     my $sth;
-    unless ($sth = $self->do_query(q{DROP INDEX IF EXISTS "%s"}, $param->{'index'})) {
+    unless ($sth =
+        $self->do_query(q{DROP INDEX IF EXISTS "%s"}, $param->{'index'})) {
         $log->syslog('err',
             'Could not drop index %s from table %s in database %s',
             $param->{'index'}, $param->{'table'}, $self->{'db_name'});
@@ -479,7 +482,7 @@ sub set_index {
             q{CREATE INDEX %s ON %s (%s)}, $param->{'index_name'},
             $param->{'table'},             $fields
         )
-        ) {
+    ) {
         $log->syslog(
             'err',
             'Could not add index %s using field %s for table %s in database %s',
@@ -489,7 +492,8 @@ sub set_index {
         );
         return undef;
     }
-    my $report = "Table $param->{'table'}, index %s set using $fields";
+    my $report = sprintf 'Table %s, index %s set using fields %s',
+        $param->{'table'}, $param->{'index_name'}, $fields;
     $log->syslog('info', 'Table %s, index %s set using fields %s',
         $param->{'table'}, $param->{'index_name'}, $fields);
     return $report;
@@ -517,9 +521,11 @@ sub translate_type {
 }
 
 # Note:
-# To prevent "database is locked" error, acquire "immediate" lock
-# by each query.  Most queries excluding "SELECT" need to lock in this
-# manner.
+# - To prevent "database is locked" error, acquire "immediate" lock
+#   by each query.  Most queries excluding "SELECT" need to lock in this
+#   manner.
+# - If a transaction has been begun, lock is not needed, because SQLite
+#   does not support nested transactions.
 sub do_query {
     my $self = shift;
     my $sth;
@@ -527,8 +533,8 @@ sub do_query {
 
     my $need_lock =
         ($_[0] =~
-            /^\s*(ALTER|CREATE|DELETE|DROP|INSERT|REINDEX|REPLACE|UPDATE)\b/i
-        );
+            /^\s*(ALTER|CREATE|DELETE|DROP|INSERT|REINDEX|REPLACE|UPDATE)\b/i)
+        unless $self->{_sdbTransactionLevel};
 
     ## acquire "immediate" lock
     unless (!$need_lock or $self->__dbh->begin_work) {
@@ -567,8 +573,8 @@ sub do_prepared_query {
 
     my $need_lock =
         ($_[0] =~
-            /^\s*(ALTER|CREATE|DELETE|DROP|INSERT|REINDEX|REPLACE|UPDATE)\b/i
-        );
+            /^\s*(ALTER|CREATE|DELETE|DROP|INSERT|REINDEX|REPLACE|UPDATE)\b/i)
+        unless $self->{_sdbTransactionLevel};
 
     ## acquire "immediate" lock
     unless (!$need_lock or $self->__dbh->begin_work) {
@@ -706,7 +712,7 @@ sub _update_table {
                     'fields'     => [sort keys %{$indexes->{$name}}]
                 }
             )
-            ) {
+        ) {
             return undef;
         }
     }
@@ -728,7 +734,7 @@ sub _get_create_table {
               WHERE type = 'table' AND name = '%s'},
             $table
         )
-        ) {
+    ) {
         $log->syslog('Could not get table \'%s\' on database \'%s\'',
             $table, $self->{'db_name'});
         return undef;
@@ -757,7 +763,7 @@ sub _copy_table {
             q{INSERT INTO "%s" (%s) SELECT %s FROM "%s"},
             $table_new, $fields, $fields, $table
         )
-        ) {
+    ) {
         $log->syslog('err',
             'Could not copy talbe "%s" to temporary table "%s_new"',
             $table, $table_new);
